@@ -23,9 +23,10 @@ THE SOFTWARE.
 """
 
 import logging
-from collections import deque, namedtuple, Counter
+from collections import namedtuple, Counter
 
 import cocotb
+from cocotb.queue import Queue
 from cocotb.triggers import Event
 
 from .version import __version__
@@ -58,21 +59,22 @@ class AxiMasterWrite(Reset):
         self.log.info("https://github.com/alexforencich/cocotbext-axi")
 
         self.aw_channel = AxiAWSource(bus.aw, clock, reset, reset_active_level)
+        self.aw_channel.queue_occupancy_limit = 2
         self.w_channel = AxiWSource(bus.w, clock, reset, reset_active_level)
+        self.w_channel.queue_occupancy_limit = 2
         self.b_channel = AxiBSink(bus.b, clock, reset, reset_active_level)
+        self.b_channel.queue_occupancy_limit = 2
 
-        self.write_command_queue = deque()
-        self.write_command_sync = Event()
-        self.write_resp_queue = deque()
-        self.write_resp_sync = Event()
+        self.write_command_queue = Queue()
+        self.current_write_command = None
 
         self.id_count = 2**len(self.aw_channel.bus.awid)
         self.cur_id = 0
         self.active_id = Counter()
 
-        self.int_write_resp_command_queue = deque()
-        self.int_write_resp_command_sync = Event()
-        self.int_write_resp_queue_list = [deque() for k in range(self.id_count)]
+        self.int_write_resp_command_queue = [Queue() for k in range(self.id_count)]
+        self.current_write_resp_command = [None for k in range(self.id_count)]
+        self.int_write_resp_queue_list = [Queue() for k in range(self.id_count)]
 
         self.in_flight_operations = 0
         self._idle = Event()
@@ -102,13 +104,17 @@ class AxiMasterWrite(Reset):
 
         self._process_write_cr = None
         self._process_write_resp_cr = None
+        self._process_write_resp_id_cr = None
 
         self._init_reset(reset, reset_active_level)
 
     def init_write(self, address, data, awid=None, burst=AxiBurstType.INCR, size=None, lock=AxiLockType.NORMAL,
             cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0, wuser=0, event=None):
 
-        if event is not None and not isinstance(event, Event):
+        if event is None:
+            event = Event()
+
+        if not isinstance(event, Event):
             raise ValueError("Expected event object")
 
         if awid is None or awid < 0:
@@ -138,8 +144,9 @@ class AxiMasterWrite(Reset):
 
         cmd = AxiWriteCmd(address, bytearray(data), awid, burst, size, lock,
             cache, prot, qos, region, user, wuser, event)
-        self.write_command_queue.append(cmd)
-        self.write_command_sync.set()
+        self.write_command_queue.put_nowait(cmd)
+
+        return event
 
     def idle(self):
         return not self.in_flight_operations
@@ -148,18 +155,9 @@ class AxiMasterWrite(Reset):
         while not self.idle():
             await self._idle.wait()
 
-    def write_resp_ready(self):
-        return bool(self.write_resp_queue)
-
-    def get_write_resp(self):
-        if self.write_resp_queue:
-            return self.write_resp_queue.popleft()
-        return None
-
     async def write(self, address, data, awid=None, burst=AxiBurstType.INCR, size=None,
             lock=AxiLockType.NORMAL, cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0, wuser=0):
-        event = Event()
-        self.init_write(address, data, awid, burst, size, lock, cache, prot, qos, region, user, wuser, event)
+        event = self.init_write(address, data, awid, burst, size, lock, cache, prot, qos, region, user, wuser)
         await event.wait()
         return event.data
 
@@ -209,39 +207,62 @@ class AxiMasterWrite(Reset):
             if self._process_write_resp_cr is not None:
                 self._process_write_resp_cr.kill()
                 self._process_write_resp_cr = None
+            if self._process_write_resp_id_cr is not None:
+                for cr in self._process_write_resp_id_cr:
+                    cr.kill()
+                self._process_write_resp_id_cr = None
+
+            self.aw_channel.clear()
+            self.w_channel.clear()
+            self.b_channel.clear()
+
+            def flush_cmd(cmd):
+                self.log.warning("Flushed write operation during reset: %s", cmd)
+                if cmd.event:
+                    cmd.event.set(None)
+
+            while not self.write_command_queue.empty():
+                cmd = self.write_command_queue.get_nowait()
+                flush_cmd(cmd)
+
+            if self.current_write_command:
+                cmd = self.current_write_command
+                self.current_write_command = None
+                flush_cmd(cmd)
+
+            for q in self.int_write_resp_command_queue:
+                while not q.empty():
+                    cmd = q.get_nowait()
+                    flush_cmd(cmd)
+
+            for k in range(len(self.current_write_resp_command)):
+                if self.current_write_resp_command[k]:
+                    cmd = self.current_write_resp_command[k]
+                    self.current_write_resp_command[k] = None
+                    flush_cmd(cmd)
+
+            for q in self.int_write_resp_queue_list:
+                while not q.empty():
+                    q.get_nowait()
+
+            self.cur_id = 0
+            self.active_id = Counter()
+
+            self.in_flight_operations = 0
+            self._idle.set()
         else:
             self.log.info("Reset de-asserted")
             if self._process_write_cr is None:
                 self._process_write_cr = cocotb.fork(self._process_write())
             if self._process_write_resp_cr is None:
                 self._process_write_resp_cr = cocotb.fork(self._process_write_resp())
-
-        self.aw_channel.clear()
-        self.w_channel.clear()
-        self.b_channel.clear()
-
-        while self.write_command_queue:
-            cmd = self.write_command_queue.popleft()
-            if cmd.event:
-                cmd.event.set(None)
-
-        while self.int_write_resp_command_queue:
-            cmd = self.int_write_resp_command_queue.popleft()
-            if cmd.event:
-                cmd.event.set(None)
-
-        self.write_resp_queue.clear()
-
-        self.in_flight_operations = 0
-        self._idle.set()
+            if self._process_write_resp_id_cr is None:
+                self._process_write_resp_id_cr = [cocotb.fork(self._process_write_resp_id(i)) for i in range(self.id_count)]
 
     async def _process_write(self):
         while True:
-            if not self.write_command_queue:
-                self.write_command_sync.clear()
-                await self.write_command_sync.wait()
-
-            cmd = self.write_command_queue.popleft()
+            cmd = await self.write_command_queue.get()
+            self.current_write_command = cmd
 
             num_bytes = 2**cmd.size
 
@@ -298,7 +319,7 @@ class AxiMasterWrite(Reset):
                     # split on 4k address boundary
                     burst_length = (min(burst_length*num_bytes, 0x1000-(cur_addr & 0xfff))+num_bytes-1)//num_bytes
 
-                    burst_list.append((awid, burst_length))
+                    burst_list.append(burst_length)
 
                     aw = self.aw_channel._transaction_obj()
                     aw.awid = awid
@@ -340,32 +361,31 @@ class AxiMasterWrite(Reset):
                 cycle_offset = (cycle_offset + num_bytes) % self.byte_width
 
             resp_cmd = AxiWriteRespCmd(cmd.address, len(cmd.data), cmd.size, cycles, cmd.prot, burst_list, cmd.event)
-            self.int_write_resp_command_queue.append(resp_cmd)
-            self.int_write_resp_command_sync.set()
+            await self.int_write_resp_command_queue[awid].put(resp_cmd)
+
+            self.current_write_command = None
 
     async def _process_write_resp(self):
         while True:
-            if not self.int_write_resp_command_queue:
-                self.int_write_resp_command_sync.clear()
-                await self.int_write_resp_command_sync.wait()
+            b = await self.b_channel.recv()
 
-            cmd = self.int_write_resp_command_queue.popleft()
+            bid = int(b.bid)
+
+            if self.active_id[bid] <= 0:
+                raise Exception(f"Unexpected burst ID {bid}")
+
+            await self.int_write_resp_queue_list[bid].put(b)
+
+    async def _process_write_resp_id(self, bid):
+        while True:
+            cmd = await self.int_write_resp_command_queue[bid].get()
+            self.current_write_resp_command[bid] = cmd
 
             resp = AxiResp.OKAY
             user = []
 
-            for bid, burst_length in cmd.burst_list:
-                while not self.int_write_resp_queue_list[bid]:
-                    b = await self.b_channel.recv()
-
-                    i = int(b.bid)
-
-                    if self.active_id[i] <= 0:
-                        raise Exception(f"Unexpected burst ID {bid}")
-
-                    self.int_write_resp_queue_list[i].append(b)
-
-                b = self.int_write_resp_queue_list[bid].popleft()
+            for burst_length in cmd.burst_list:
+                b = await self.int_write_resp_queue_list[bid].get()
 
                 burst_id = int(b.bid)
                 burst_resp = AxiResp(b.bresp)
@@ -389,11 +409,9 @@ class AxiMasterWrite(Reset):
 
             write_resp = AxiWriteResp(cmd.address, cmd.length, resp, user)
 
-            if cmd.event is not None:
-                cmd.event.set(write_resp)
-            else:
-                self.write_resp_queue.append(write_resp)
-                self.write_resp_sync.set()
+            cmd.event.set(write_resp)
+
+            self.current_write_resp_command[bid] = None
 
             self.in_flight_operations -= 1
 
@@ -411,20 +429,20 @@ class AxiMasterRead(Reset):
         self.log.info("https://github.com/alexforencich/cocotbext-axi")
 
         self.ar_channel = AxiARSource(bus.ar, clock, reset, reset_active_level)
+        self.ar_channel.queue_occupancy_limit = 2
         self.r_channel = AxiRSink(bus.r, clock, reset, reset_active_level)
+        self.r_channel.queue_occupancy_limit = 2
 
-        self.read_command_queue = deque()
-        self.read_command_sync = Event()
-        self.read_data_queue = deque()
-        self.read_data_sync = Event()
+        self.read_command_queue = Queue()
+        self.current_read_command = None
 
         self.id_count = 2**len(self.ar_channel.bus.arid)
         self.cur_id = 0
         self.active_id = Counter()
 
-        self.int_read_resp_command_queue = deque()
-        self.int_read_resp_command_sync = Event()
-        self.int_read_resp_queue_list = [deque() for k in range(self.id_count)]
+        self.int_read_resp_command_queue = [Queue() for k in range(self.id_count)]
+        self.current_read_resp_command = [None for k in range(self.id_count)]
+        self.int_read_resp_queue_list = [Queue() for k in range(self.id_count)]
 
         self.in_flight_operations = 0
         self._idle = Event()
@@ -452,13 +470,17 @@ class AxiMasterRead(Reset):
 
         self._process_read_cr = None
         self._process_read_resp_cr = None
+        self._process_read_resp_id_cr = None
 
         self._init_reset(reset, reset_active_level)
 
     def init_read(self, address, length, arid=None, burst=AxiBurstType.INCR, size=None,
             lock=AxiLockType.NORMAL, cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0, event=None):
 
-        if event is not None and not isinstance(event, Event):
+        if event is None:
+            event = Event()
+
+        if not isinstance(event, Event):
             raise ValueError("Expected event object")
 
         if length < 0:
@@ -483,8 +505,9 @@ class AxiMasterRead(Reset):
         self._idle.clear()
 
         cmd = AxiReadCmd(address, length, arid, burst, size, lock, cache, prot, qos, region, user, event)
-        self.read_command_queue.append(cmd)
-        self.read_command_sync.set()
+        self.read_command_queue.put_nowait(cmd)
+
+        return event
 
     def idle(self):
         return not self.in_flight_operations
@@ -493,18 +516,9 @@ class AxiMasterRead(Reset):
         while not self.idle():
             await self._idle.wait()
 
-    def read_data_ready(self):
-        return bool(self.read_data_queue)
-
-    def get_read_data(self):
-        if self.read_data_queue:
-            return self.read_data_queue.popleft()
-        return None
-
     async def read(self, address, length, arid=None, burst=AxiBurstType.INCR, size=None,
             lock=AxiLockType.NORMAL, cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0):
-        event = Event()
-        self.init_read(address, length, arid, burst, size, lock, cache, prot, qos, region, user, event)
+        event = self.init_read(address, length, arid, burst, size, lock, cache, prot, qos, region, user)
         await event.wait()
         return event.data
 
@@ -554,38 +568,61 @@ class AxiMasterRead(Reset):
             if self._process_read_resp_cr is not None:
                 self._process_read_resp_cr.kill()
                 self._process_read_resp_cr = None
+            if self._process_read_resp_id_cr is not None:
+                for cr in self._process_read_resp_id_cr:
+                    cr.kill()
+                self._process_read_resp_id_cr = None
+
+            self.ar_channel.clear()
+            self.r_channel.clear()
+
+            def flush_cmd(cmd):
+                self.log.warning("Flushed read operation during reset: %s", cmd)
+                if cmd.event:
+                    cmd.event.set(None)
+
+            while not self.read_command_queue.empty():
+                cmd = self.read_command_queue.get_nowait()
+                flush_cmd(cmd)
+
+            if self.current_read_command:
+                cmd = self.current_read_command
+                self.current_read_command = None
+                flush_cmd(cmd)
+
+            for q in self.int_read_resp_command_queue:
+                while not q.empty():
+                    cmd = q.get_nowait()
+                    flush_cmd(cmd)
+
+            for k in range(len(self.current_read_resp_command)):
+                if self.current_read_resp_command[k]:
+                    cmd = self.current_read_resp_command[k]
+                    self.current_read_resp_command[k] = None
+                    flush_cmd(cmd)
+
+            for q in self.int_read_resp_queue_list:
+                while not q.empty():
+                    q.get_nowait()
+
+            self.cur_id = 0
+            self.active_id = Counter()
+
+            self.in_flight_operations = 0
+            self._idle.set()
         else:
             self.log.info("Reset de-asserted")
             if self._process_read_cr is None:
                 self._process_read_cr = cocotb.fork(self._process_read())
             if self._process_read_resp_cr is None:
                 self._process_read_resp_cr = cocotb.fork(self._process_read_resp())
-
-        self.ar_channel.clear()
-        self.r_channel.clear()
-
-        while self.read_command_queue:
-            cmd = self.read_command_queue.popleft()
-            if cmd.event:
-                cmd.event.set(None)
-
-        while self.int_read_resp_command_queue:
-            cmd = self.int_read_resp_command_queue.popleft()
-            if cmd.event:
-                cmd.event.set(None)
-
-        self.read_data_queue.clear()
-
-        self.in_flight_operations = 0
-        self._idle.set()
+            if self._process_read_resp_id_cr is None:
+                self._process_read_resp_id_cr = [cocotb.fork(self._process_read_resp_id(i)) for i in range(self.id_count)]
 
     async def _process_read(self):
         while True:
-            if not self.read_command_queue:
-                self.read_command_sync.clear()
-                await self.read_command_sync.wait()
-
-            cmd = self.read_command_queue.popleft()
+            cmd = await self.read_command_queue.get()
+            self.current_read_command = cmd
 
             num_bytes = 2**cmd.size
 
@@ -619,7 +656,7 @@ class AxiMasterRead(Reset):
                     # split on 4k address boundary
                     burst_length = (min(burst_length*num_bytes, 0x1000-(cur_addr & 0xfff))+num_bytes-1)//num_bytes
 
-                    burst_list.append((arid, burst_length))
+                    burst_list.append(burst_length)
 
                     ar = self.r_channel._transaction_obj()
                     ar.arid = arid
@@ -643,16 +680,25 @@ class AxiMasterRead(Reset):
                 cur_addr += num_bytes
 
             resp_cmd = AxiReadRespCmd(cmd.address, cmd.length, cmd.size, cycles, cmd.prot, burst_list, cmd.event)
-            self.int_read_resp_command_queue.append(resp_cmd)
-            self.int_read_resp_command_sync.set()
+            await self.int_read_resp_command_queue[arid].put(resp_cmd)
+
+            self.current_read_command = None
 
     async def _process_read_resp(self):
         while True:
-            if not self.int_read_resp_command_queue:
-                self.int_read_resp_command_sync.clear()
-                await self.int_read_resp_command_sync.wait()
+            r = await self.r_channel.recv()
 
-            cmd = self.int_read_resp_command_queue.popleft()
+            rid = int(r.rid)
+
+            if self.active_id[rid] <= 0:
+                raise Exception(f"Unexpected burst ID {rid}")
+
+            await self.int_read_resp_queue_list[rid].put(r)
+
+    async def _process_read_resp_id(self, rid):
+        while True:
+            cmd = await self.int_read_resp_command_queue[rid].get()
+            self.current_read_resp_command[rid] = cmd
 
             num_bytes = 2**cmd.size
 
@@ -669,19 +715,9 @@ class AxiMasterRead(Reset):
 
             first = True
 
-            for rid, burst_length in cmd.burst_list:
+            for burst_length in cmd.burst_list:
                 for k in range(burst_length):
-                    while not self.int_read_resp_queue_list[rid]:
-                        r = await self.r_channel.recv()
-
-                        i = int(r.rid)
-
-                        if self.active_id[i] <= 0:
-                            raise Exception(f"Unexpected burst ID {rid}")
-
-                        self.int_read_resp_queue_list[i].append(r)
-
-                    r = self.int_read_resp_queue_list[rid].popleft()
+                    r = await self.int_read_resp_queue_list[rid].get()
 
                     cycle_id = int(r.rid)
                     cycle_data = int(r.rdata)
@@ -724,11 +760,9 @@ class AxiMasterRead(Reset):
 
             read_resp = AxiReadResp(cmd.address, data, resp, user)
 
-            if cmd.event is not None:
-                cmd.event.set(read_resp)
-            else:
-                self.read_data_queue.append(read_resp)
-                self.read_data_sync.set()
+            cmd.event.set(read_resp)
+
+            self.current_read_resp_command[rid] = None
 
             self.in_flight_operations -= 1
 
@@ -746,11 +780,11 @@ class AxiMaster:
 
     def init_read(self, address, length, arid=None, burst=AxiBurstType.INCR, size=None,
             lock=AxiLockType.NORMAL, cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0, event=None):
-        self.read_if.init_read(address, length, arid, burst, size, lock, cache, prot, qos, region, user, event)
+        return self.read_if.init_read(address, length, arid, burst, size, lock, cache, prot, qos, region, user, event)
 
     def init_write(self, address, data, awid=None, burst=AxiBurstType.INCR, size=None, lock=AxiLockType.NORMAL,
             cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0, wuser=0, event=None):
-        self.write_if.init_write(address, data, awid, burst, size, lock, cache, prot, qos, region, user, wuser, event)
+        return self.write_if.init_write(address, data, awid, burst, size, lock, cache, prot, qos, region, user, wuser, event)
 
     def idle(self):
         return (not self.read_if or self.read_if.idle()) and (not self.write_if or self.write_if.idle())
@@ -766,18 +800,6 @@ class AxiMaster:
     async def wait_write(self):
         await self.write_if.wait()
 
-    def read_data_ready(self):
-        return self.read_if.read_data_ready()
-
-    def get_read_data(self):
-        return self.read_if.get_read_data()
-
-    def write_resp_ready(self):
-        return self.write_if.write_resp_ready()
-
-    def get_write_resp(self):
-        return self.write_if.get_write_resp()
-
     async def read(self, address, length, arid=None, burst=AxiBurstType.INCR, size=None,
             lock=AxiLockType.NORMAL, cache=0b0011, prot=AxiProt.NONSECURE, qos=0, region=0, user=0):
         return await self.read_if.read(address, length, arid,
